"""
a tiny tiny blog.  
this is the view class and is more bitsyblog than anything
else can claim to be
"""

### global variables

# format to uniquely label blog posts
timestamp = '%Y%m%d%H%M%S' # strftime representation

# who can view which blog posts
roles = { 'public': ( 'public', ),
          'friend': ( 'public', 'secret' ),
          'author': ( 'public', 'secret', 'private' ), }

### imports

import dateutil.parser  # XXX separate, for now
import parser # bitsyblog dateutil parser

import cgi
import datetime
import docutils
import docutils.core
import inspect
import markup
import os
import re
import utils

from blog import FileBlog
from user import FilespaceUsers
from lxml import etree
from markup.form import Form
from cStringIO import StringIO
from webob import Request, Response, exc

### exceptions

class DateStampException(Exception):
    """exception when parsing a datestamp"""
    # probably don't need this anymore

class BlogPathException(Exception):
    """exception when trying to retrieve the blog"""

### the main course

class BitsyBlog(object):
    """a very tiny blog"""

    ### class level variables
    defaults = { 'date_format': '%H:%M %F',
                 'file_dir': os.path.dirname(__file__),
                 'subject': '[ %(date)s ]:',
                 'n_links': 5, # number of links for navigation
                 'site_name': 'bitsyblog'
                 }

    def __init__(self, **kw):
        for key in self.defaults:
            setattr(self, key, kw.get(key, self.defaults[key]))
        self.n_links = int(self.n_links) # could be a string from the .ini
        self.response_functions = { 'GET': self.get,
                                    'POST': self.post,
                                    'PUT': self.put
                                    }
        
        # abstract attributes
        self.users = FilespaceUsers(self.file_dir)
        self.blog = FileBlog(self.file_dir)
        self.cooker = self.restructuredText

        # for BitsyAuth (for now)
        self.passwords = self.users.passwords
        self.newuser = self.users.new

    ### methods dealing with HTTP

    def __call__(self, environ, start_response):
        self.request = Request(environ)
        res = self.make_response(self.request.method)
        return res(environ, start_response)

    def make_response(self, method):
        return self.response_functions.get(method, self.error)()

    def get_response(self, text, content_type='text/html'):
        res = Response(content_type=content_type, body=text)
        res.content_length = len(res.body)
        return res

    def get(self):
        """
        display the blog or respond to a get request
        """

        # determine number of navigation links to display
        n_links = self.request.GET.get('n')
        if n_links == 'all':
            n_links = -1
        try:
            n_links = int(n_links)
        except (TypeError, ValueError):
            n_links = self.n_links

        path = self.request.path_info.strip('/')
        if not path: # the front page
            return self.get_response(self.index(n_links))

        ### static files

        # site.css 
        if path == 'css/site.css':
            css_file = os.path.join(self.file_dir, 'site.css')
            return self.get_response(file(css_file).read(), content_type='text/css')

        # logo
        if path == 'bitsyblog.png':
            logo = os.path.join(self.file_dir, 'bitsyblog.png')
            return self.get_response(file(logo, 'rb').read(), content_type='image/png')

        ### user space

        user, path = self.user()
        if user not in self.users:
            return exc.HTTPNotFound("No blog found for %s" % user)            
        self.request.user = self.users[user]

        check = self.check_user(user)

        # special paths
        if path == [ 'post' ]:
            if check is not None:
                return check
            return self.get_response(self.form_post(user))

        if path == [ 'preferences' ]:

            if check is not None:
                return check
            return self.get_response(self.preferences(user))

        if len(path) == 2:
            if path[0] == 'css':
                for i in self.request.user.settings['CSS']:
                    # find the right CSS file
                    if i['filename'] == path[1]:
                        return self.get_response(i['css'], content_type='text/css')
                else:
                    return exc.HTTPNotFound('CSS "%s" not found' % path[1])

        role = self.role(user)

        # get the blog
        try:
            blog = self.get_blog(user, path, role)
        except BlogPathException, e:
            return exc.HTTPNotFound(str(e))
        except exc.HTTPException, e:
            return e.wsgi_response

        # don't display navigation for short blogs
        if len(blog) < 2:
            n_links = 0

        # reverse the blog if necessary
        if self.request.GET.get('sort') == 'forward':
            blog = list(reversed(blog))
            
        # write the blog
        content = self.write_blog(user, blog, self.request.path_info, n_links)

        # return the content
        return self.get_response(content)

    def post(self):
        """
        write a blog entry and other POST requests
        """
        # TODO: posting to '/' ?
        
        # find user + path
        user, path = self.user()

        if user not in self.users:
            return exc.HTTPNotFound("No blog found for %s" % user)
        self.request.user = self.users[user]

        check = self.check_user(user)
        if check is not None:
            return check

        if len(path):
            if path == [ 'preferences' ]:
            
                # make the data look like we want
                settings = {}
                settings['Date format'] = self.request.POST.get('Date format')
                settings['Subject'] = '%(date)s'.join((self.request.POST['Subject-0'], self.request.POST['Subject-2']))
                settings['Stylesheet'] = self.request.POST['Stylesheet']
                settings['CSS file'] = self.request.POST.get('CSS file')
                settings['Friends'] = ', '.join(self.request.POST.getall('Friends'))
                
                errors = self.users.write_settings(user, **settings)
                if errors: # re-display form with errors                    
                    return self.get_response(self.preferences(user, errors))
                
                return self.get_response(self.preferences(user, message='Changes saved'))
            elif len(path) == 1 and self.isentry(path[0]):
                entry = self.blog.entry(path[0])
                if entry is None:
                    return exc.HTTPNotFound("Blog entry %s not found %s" % path[0])
                
            else:
                return exc.HTTPMethodNotAllowed("Not sure what you're trying to do")
        
        # get the body of the post
        body = self.request.body
        body = self.request.POST.get('form-post', body)
        body = body.strip()
        if not body:
            pass # do something

        # determine if the post is secret or private
        privacy = self.request.GET.get('privacy') or self.request.POST.get('privacy') or 'public'

        # write the file
        now = self.datestamp(datetime.datetime.now())
        location = "/%s/%s" % ( user, now )
        self.blog.post(user, now, body, privacy)

        # point the user at the post
        return exc.HTTPSeeOther("Post blogged by bitsy", location=location)

    def put(self):
        """
        PUT several blog entries from a file
        """

        # find user + path
        user, path = self.user()

        if user not in self.users.users():
            return exc.HTTPNotFound("No blog found for %s" % user)

        if len(path):
            return exc.HTTPMethodNotAllowed("Not sure what you're trying to do")
        
        # find the dates + entries in the file
        regex = '\[.*\]:'
        entries = re.split(regex, self.request.body)[1:]
        dates = [ date.strip().strip(':').strip('[]').strip()
                  for date in re.findall(regex, self.request.body) ]
        dates = [ dateutil.parser.parse(date) for date in dates ]

        # write to the blog
        for i in range(len(entries)):
            datestamp = datetime.strftime(timestamp)
            self.blog.post(user, datestamp, 'public')
        
        return exc.HTTPOk("%s posts blogged by bitsy" % len(entries))


    def error(self):
        """deal with non-supported methods"""
        methods = ', '.join(self.response_functions.keys()[:1]) 
        methods += ' and %s' % self.response_functions.keys()[-1] 
        return exc.HTTPMethodNotAllowed("Only %s operations are allowed" % methods)

    ### auth functions

    def authenticated(self):
        """return authenticated user"""
        return self.request.environ.get('REMOTE_USER')

    def check_user(self, user):
        """
        determine authenticated user
        returns None on success
        """
        authenticated = self.authenticated()
        if authenticated is None:
            return exc.HTTPUnauthorized('Unauthorized')
        elif user != authenticated:
            return exc.HTTPForbidden("Forbidden")        

    def role(self, user):
        """
        determine what role the authenticated member has
        with respect to the user
        """
        auth = self.authenticated()
        if not auth:
            return 'public'
        if auth == user:
            return 'author'
        else:
            if auth in self.request.user.settings['Friends']:
                return 'friend'
            else:
                return 'public'

    ### user methods

    def user(self): # TODO -> def userpath(self)
        """user who's blog one is viewing"""        
        path = self.request.path_info.strip('/').split('/')
        name = path[0]
        path = path[1:]
        if not name:
            name = None
        return name, path

    ### date methods

    def datestamp(self, date):
        # TODO -> utils.py
        return date.strftime(timestamp)

    def isentry(self, string):
        return (len(string) == len(''.join(utils.timeformat))) and string.isdigit()

    ### blog retrival methods

    def get_blog(self, user, path, role='public'):
        """retrieve the blog entry based on the path"""

        # get permissions
        allowed = roles[role]

        # entire blog
        if not path:
            return self.blog(user, allowed)

        # individual blog entry
        if (len(path) == 1) and self.isentry(path[0]):
            blog = self.blog.entry(user, path[0], allowed)
            if not blog:
                raise BlogPathException("No blog entry for %s" % path[0])
            return [ blog ]
            
        # parse the path into a date path
        n_date_vals = 3 # year, month, date
        if len(path) > n_date_vals:
            raise BlogPathException("blog entry not found")

        # ensure the path conforms to expected values (ints):
        try:
            [ int(i) for i in path]
        except ValueError:
            raise BlogPathException("blog entry not found")

        # get the blog
        return self.blog.entries(user, allowed, *path)

    ### methods that write HTML

    def render(self, body, title=None):
        """layout the page in a unified way"""
        stylesheets = ()
        user = getattr(self.request, 'user', None)
        _title = [ self.site_name ]
        if user:
            stylesheets = self.request.user['CSS']
            stylesheets = [ (("/%s/css/%s" % (user.name, css['filename'])),
                             css['name']) for css in stylesheets ]
            _title.insert(0, self.request.user.name)
        else:
            stylesheets = (("/css/site.css", "Default"),)
            
        if title:
            _title.insert(0, title)
            
        title = ' - '.join(_title)
        return markup.wrap(self.site_nav()+body, title, stylesheets)

    def site_nav(self):
        """returns HTML for site navigation"""
        links = [('/',), ]
        user = self.authenticated()
        if user:
            links.extend([('/%s' % user, user),
                          ('/%s/post' % user, 'post'),
                          ('/%s/preferences' % user, 'preferences'),
                          ('/logout', 'logout')])
        else:
            links.extend([('/login', 'login'), ('/join', 'join')])
        links = [ markup.link(*i) for i in links ]
        return markup.listify(links, ordered=False, **{ 'class': 'site-nav'})

    def index(self, n_links):
        retval = StringIO()
        print >> retval, '<h1><img src="bitsyblog.png" alt="bitsyblog"/></h1>'

        # get the blogs
        blogs = {}
        for user in self.users:
            blog = self.blog(user, ('public',), n_links)
            if blog:
                blogs[user] = blog
        users = blogs.keys()

        # display latest active user first
        users.sort(key=lambda user: blogs[user][0].date, reverse=True)

        # display users' blogs
        for user in users:
            print >> retval, '<div id="%s" class="user">' % user
            print >> retval, '<a href="%s">%s</a>' % (user, user)
            blog = blogs[user]
            print >> retval, self.navigation(user, blog, '/%s' % user, n_links)
            print >> retval, '</div>'

        return self.render(retval.getvalue())

    def navigation(self, user, blog, path, n_links, n_char=80):
        prefs = self.users[user].settings
        
        if n_links == 0 or not len(blog):
            return ''
        retval = StringIO()
        print >> retval, '<div class="navigation">'
        more = ''
        if (n_links != -1) and (len(blog) > n_links):
            more = '<a href="%s?n=all">more</a>' % path
            blog = blog[:n_links]

        entries = []
        for entry in blog:
            id = self.datestamp(entry.date)
            format = prefs.get('Date format', self.date_format)
            datestamp = entry.date.strftime(format)
            synopsis = entry.snippet()
            if synopsis:
                synopsis = ': %s' % cgi.escape(synopsis)
            entries.append(markup.link("%s#%s" % (path, id), datestamp) + synopsis)

        print >> retval, markup.listify(entries)
        print >> retval, more
        print >> retval, '</div>'
        return retval.getvalue()

    def blog_entry(self, user, entry):
        """given the content string, return a marked-up blog entry"""

        # user preferences
        prefs = self.request.user.settings
        format = prefs.get('Date format', self.date_format)
        subject = prefs.get('Subject', self.subject)

        role = self.role(user)
        
        subject = subject % { 'date' : entry.date.strftime(format) }
        subject = cgi.escape(subject)
        html = StringIO()
        id = self.datestamp(entry.date)
        print >> html, '<div id="%s" class="blog-entry">' % id
        print >> html, '<a name="%s" />' % id
        print >> html, '<div class="subject">'
        print >> html, '<a href="/%s/%s">%s</a>' % (user, id, subject)
        if (entry.privacy == 'secret') and (role == 'friend'):
            print >> html, '<em>secret</em>'
        print >> html, '</div>'
        print >> html, self.cooker(entry.body)

        if role == 'author':
            print >> html, '<div><form>'
            print >> html, self.privacy_settings(entry.privacy)
            print >> html, '</form></div>'
        
        print >> html, '</div>'
        return html.getvalue()
        
    def write_blog(self, user, blog, path, n_links):
        """return the user's blog in HTML"""
        retval = StringIO()
        print >> retval,  self.navigation(user, blog, path, n_links, 0)
        for entry in blog:
            print >> retval, self.blog_entry(user, entry)
        return self.render(retval.getvalue())
                           
    def restructuredText(self, string):
        origstring = string
        settings = { 'report_level': 5 }
        parts = docutils.core.publish_parts(string.strip(),
                                            writer_name='html',
                                            settings_overrides=settings)
        retval = '<div class="blog-body">%s</div>' % parts['body']
        try:
            foo = etree.fromstring(retval)
        except etree.XMLSyntaxError:
            pass
        # should cleanup the <div class="system-message">
        for i in foo.getiterator():
            if dict(i.items()).get('class') ==  'system-message':
                i.clear()
                
        return etree.tostring(foo)

    ### forms and accompanying display

    def form_post(self, user):
        retval = StringIO()
        print >> retval, '<form action="/%s" method="post">' % user
        print >> retval, '<textarea cols="80" rows="25" name="form-post"></textarea><br/>'
        print >> retval, self.privacy_settings()
        print >> retval, '<input type="submit" name="submit" value="Post" />'
        print >> retval, '</form>'
        return self.render(retval.getvalue())

    def preferences_form(self, user):
        prefs = self.request.user.settings
        form = Form()

        # date format
        format = prefs.get('Date format', self.date_format)
        value = datetime.datetime.now().strftime(format)
        form.add_element('textfield', 'Date format', value=value,
                         help='how to display dates in your blog post subject')

        # subject
        subject = prefs.get('Subject', self.subject)
        subject = subject.split('%(date)s', 1)
        func = lambda name: value
        form.add_elements('Subject',
                          ['textfield', func, 'textfield' ],
                          [ (), (), ()], # this is horrible!
                          [ dict(value=subject[0]),
                            {},
                            dict(value=subject[1]) ],
                          help='how to display the subject line of your blog post'
                          )
                          
        # CSS files
        css_files = [ i['name'] for i in prefs['CSS'] ]
        form.add_element('menu', 'Stylesheet', css_files,
                         help='which CSS file should be the default')

        # or upload a CSS file
        form.add_element('file_upload', 'CSS file',
                         help='upload a CSS file to theme your webpage')

        # Friends -- can see secret posts
        users = [ i for i in list(self.users.users()) 
                  if i != user ]
        if users:
            users.sort()
            form.add_element('checkboxes', 'Friends',
                             users, prefs.get('Friends', set()),
                             help='friends can see your secret posts')
        
        return form

    def preferences(self, user, errors=None, message=None):
        """user preferences form"""
        body = self.preferences_form(user)(errors)
        if message:
            body = '%s\n%s' % ( markup.p(markup.strong(message)), body )
        return self.render(body, title='preferences')

    def privacy_settings(self, default='public'):
        """HTML snippet for privacy settings"""
        settings = (('public', 'viewable to everyone'),
                    ('secret', 'viewable only to your friends'),
                    ('private', 'viewable only to you'))
        form = Form()
        return form.radiobuttons('privacy', settings, checked=default, joiner=' ')



#!/usr/bin/env python

import optparse
import os
import subprocess
import sys
import tempfile
import urllib2

# global variables

EDITOR='emacs -nw'
SERVER='http://bitsyblog.biz'
dotfile='.blogme'

parser = optparse.OptionParser()
parser.add_option('-s', '--server', default=SERVER)
parser.add_option('-u', '--user')
parser.add_option('-p', '--password')
parser.add_option('--private', action='store_true', default=False)
parser.add_option('--secret', action='store_true', default=False)

options, args = parser.parse_args()

if options.private and option.secret:
    print "post can't be secret and private!"
    sys.exit(1)

# parse dotfile

home = os.environ.get('HOME')
if home:
    dotfile = os.path.join(home, dotfile)
    if os.path.exists(dotfile):
        prefs = file(dotfile).read().split('\n')
        prefs = [ i for i in prefs if i.strip() ]
        prefs = [ [ j.strip() for j in i.split(':', 1) ] for i in prefs
                  if ':' in i] # probably not necessary
        prefs = dict(prefs)
    else:
        prefs = {}

# determine user name and password
fields = [ 'user', 'password' ]
for field in fields:
    globals()[field] = prefs.get(field)

    optval = getattr(options, field)
    if optval:
        password = None # needed to ensure prompting for pw from command line
        globals()[field] = optval

    if globals()[field] is None:
        globals()[field] = raw_input('%s: ' % field)
assert user is not None
assert password is not None

# write the dotfile if it doesn't exist
if not os.path.exists(dotfile):
    preffile = file(dotfile, 'w')
    print >> preffile, 'user: %s' % user
    print >> preffile, 'password: %s' % password
    preffile.close()
    os.chmod(dotfile, 0600)
        
def tmpbuffer(editor=EDITOR):
    """open an editor and retreive the resulting editted buffer"""
    tmpfile = tempfile.mktemp(suffix='.txt')
    cmdline = editor.split()
    cmdline.append(tmpfile)
    edit = subprocess.call(cmdline)
    buffer = file(tmpfile).read().strip()
    os.remove(tmpfile)
    return buffer

# get the blog

if args:
    msg = ' '.join(args)
else:
    msg = tmpbuffer()

# open the url

url = '/'.join((options.server, user))
url += '?auth=digest' # specify authentication method

if options.private:
    url += '&privacy=private'
if options.secret:
    url += '&privacy=secret'

authhandler = urllib2.HTTPDigestAuthHandler()
authhandler.add_password('bitsyblog', url, user, password)
opener = urllib2.build_opener(authhandler)
urllib2.install_opener(opener)
url = urllib2.urlopen(url, data=msg)

# print the blog post's url

print url.url



from bitsyauth import BitsyAuth
from bitsyblog import BitsyBlog
from paste.httpexceptions import HTTPExceptionHandler

def factory(global_conf, **app_conf):
    """make bitsyauth app and wrap it in middleware"""
    
    config = [ 'file_dir', 'date_format', 'subject', 'n_links' ]
    key_str = 'bitsyblog.%s'
    args = dict([ (key, app_conf[ key_str % key]) for key in config
                  if app_conf.has_key(key_str % key) ])

    app = BitsyBlog(**args)
    secret = app_conf.get('secret', 'secret')
    return BitsyAuth(HTTPExceptionHandler(app), global_conf, app.passwords, app.newuser, 'bitsyblog', secret)

# TODO: use wsgifilter.proxyapp.DebugHeaders to debug the headers apache
# doesn't like



import markup
import random
import re
import sys

from cStringIO import StringIO
from markup.form import Form
from paste.auth import basic, cookie, digest, form, multi, auth_tkt
from webob import Request, Response, exc

try:
    from skimpyGimpy import skimpyAPI
    CAPTCHA = True
except ImportError:
    CAPTCHA = False

dictionary_file = '/usr/share/dict/american-english'

def random_word():
    """generate a random word for CAPTCHA auth"""
    min_length = 5 # minimum word length
    if not globals().has_key('dictionary'):
        # read the dictionary -- this may be platform dependent
        _dictionary = file(dictionary_file).readlines()
        _dictionary = [ i.strip() for i in _dictionary ]
        _dictionary = [ i.lower() for i in _dictionary
                        if i.isalpha() and i > min_length ]
        globals()['dictionary'] = _dictionary
    return random.Random().choice(dictionary)

class BitsyAuthInnerWare(object):
    """inner auth;  does login checking"""

    def __init__(self, app, passwords, newuser=None, site=None, realm=None):
        """a simple reimplementation of auth
        * app: the WSGI app to be wrapped
        * passwords: callable that return a dictionary of {'user': 'password'}
        * newuser: callable to make a new user, taking name + pw
        * site: name of the site
        * realm: realm for HTTP digest authentication
        """

        self.app = app
        self.passwords = passwords
        self.site = site or ''
        self.realm = realm or self.site
        self.captcha = True
        self.redirect_to = '/' # redirect to site root        
        self.urls = { 'login': '/login', 'join': '/join', }
        self.keys = {} # keys, words for CAPTCHA request

        self.content_type = { 'image_captcha': 'image/png',
                              'wav_captcha': 'audio/wav' }

        # WSGI app securely wrapped
        self.wrapped_app = self.security_wrapper()

        if newuser:
            self.newuser = newuser
        else:
            self.urls.pop('join') # don't do joining

        if not CAPTCHA:
            self.captcha = False
        
    ### WSGI/HTTP layer

    def __call__(self, environ, start_response):

        self.request = Request(environ)
        self.request.path_info = self.request.path_info.rstrip('/')

        # URLs intrinsic to BitsyAuth
        if self.request.path_info == '/logout':
            response = self.redirect()
            return response(self.request.environ, start_response)

        if self.request.path_info in self.url_lookup():
            response = self.make_response()
            return response(self.request.environ, start_response)

        # digest auth
        if self.request.headers.has_key('Authorization'):
            return self.wrapped_app(self.request.environ, start_response)

        response = self.request.get_response(self.app)
        # respond to 401s
        if response.status_int == 401: # Unauthorized
            if self.request.environ.get('REMOTE_USER'):
                return exc.HTTPForbidden()
            else:
                response = self.request.get_response(self.wrapped_app)

        user = self.request.environ.get('REMOTE_USER')
        if user:
            self.request.environ['paste.auth_tkt.set_user'](user)

        return response(self.request.environ, start_response)

    ### authentication function

    def digest_authfunc(self, environ, realm, user):
        return self.passwords()[user]
        digested = digest.digest_password(realm, user, self.passwords()[user])
        return digested

    def authfunc(self, environ, user, password):
        return self.hash(user, password) == self.passwords()[user]

    def hash(self, user, password):
        # use md5 digest for now
        return digest.digest_password(self.realm, user, password)

    def security_wrapper(self):
        """return the app securely wrapped"""

        multi_auth = multi.MultiHandler(self.app)

        # digest authentication
        multi_auth.add_method('digest', digest.middleware,
                              self.realm, self.digest_authfunc)
        multi_auth.set_query_argument('digest', key='auth')

        # form authentication
        template = self.login(wrap=True, action='%s')
        multi_auth.add_method('form', form.middleware, self.authfunc,
                              template=template)
        multi_auth.set_default('form')

        return multi_auth

        # might have to wrap cookie.middleware(BitsyAuth(multi(app))) ::shrug::
        return cookie.middleware(multi_auth)

    ### methods dealing with intrinsic URLs

    def url_lookup(self):
        retval = dict([ (value, key) for key, value
                        in self.urls.items() ])
        if self.captcha:
            retval.update(dict([(('/join/%s.png' % key), 'image_captcha')
                                for key in self.keys]))
        return retval
        
    def get_response(self, text, content_type='text/html'):
        res = Response(content_type=content_type, body=text)
        res.content_length = len(res.body)
        return res

    def make_response(self):
        url_lookup = self.url_lookup()
        path = self.request.path_info
        assert path in url_lookup

        # login and join shouldn't be accessible when logged in
        if self.request.environ.get('REMOTE_USER'):
            return self.redirect("You are already logged in")

        handler = url_lookup[path]
        function = getattr(self, handler)

        if self.request.method == 'GET':
            # XXX could/should do this with decorators            
            return self.get_response(function(wrap=True),
                                     content_type=self.content_type.get(handler,'text/html'))
        if self.request.method == 'POST':
            post_func = getattr(self, handler + "_post")
            errors = post_func()
            if errors:
                return self.get_response(function(errors=errors, wrap=True))
            else:
                return self.redirect("Welcome!")

    def redirect(self, message=''):
        """redirect from instrinsic urls"""
        return exc.HTTPSeeOther(message, location=self.redirect_to)

    def image_captcha(self, wrap=True):
        """return data for the image"""
        key = self.request.path_info.split('/join/')[-1]
        key = int(key.split('.png')[0])
        return skimpyAPI.Png(self.keys[key], scale=3.0).data()
                
    ### forms and their display methods

    ### login

    def login_form(self, referer=None, action=None):
        if action is None:
            action = self.urls['login']
        form = Form(action=action, submit='Login')
        form.add_element('textfield', 'Name', input_name='username')
        form.add_element('password', 'Password', input_name='password')
        if referer is not None:
            form.add_element('hidden', 'referer', value=referer)
        return form

    def login(self, errors=None, wrap=False, action=None):
        """login div"""
        form = self.login_form(action=action)
        join = self.urls.get('join')
        retval = form(errors)
        if join:        
            retval += '<br/>\n' + markup.a('join', href="%s" % join)
        retval = markup.div(retval)
        if wrap:
            title = 'login'
            if self.site:
                pagetitle = '%s - %s' % (title, self.site)
            retval = markup.wrap(markup.h1(title.title()) + retval,
                                 pagetitle=pagetitle)

        return retval

    def login_post(self):
        """handle a login POST request"""
        user = self.request.POST.get('username')
        password = self.request.POST.get('password')
        passwords = self.passwords()
        error = False
        if user not in passwords:
            error = True
        else:
            error = not self.authfunc(self.request.environ, user, password)
        if error:
            return { 'Name': 'Wrong username or password' }
        self.request.environ['REMOTE_USER'] = user
        self.request.environ['paste.auth_tkt.set_user'](user)

    ### join

    def captcha_pre(self, word, key):
        """CAPTCHA with pre-formatted text"""
        return skimpyAPI.Pre(word, scale=1.2).data()

    def captcha_png(self, word, key):
        """CAPTCHA with a PNG image"""
        return markup.image('/join/%s.png' % key)

    def join_form(self):
        captcha = ''
        if self.captcha:
            # data for CAPTCHA
            key = random.Random().randint(0, sys.maxint)
            word = random_word()

            self.keys[key] = word

            captcha = StringIO()

            captcha_text = "Please type the word below so I know you're not a computer:"
            captcha_help = "(please %s if the page is unreadable)" % markup.link('/join?captcha=image', 'go here')

            print >> captcha, markup.p('%s<br/> %s' % (captcha_text, 
                                                       markup.i(captcha_help)))

            # determine type of CAPTCHA
            captchas = ' '.join(self.request.GET.getall('captcha'))
            if not captchas:
                captchas = 'pre'
                
            captcha_funcs=dict(pre=self.captcha_pre,
                               image=self.captcha_png,)
            captchas = [ captcha_funcs[i](word, key) for i in captchas.split()
                         if i in captcha_funcs ]
            captchas = '\n'.join([markup.p(i) for i in captchas])
            print >> captcha, captchas
            
            print >> captcha, markup.p(markup.input(None, **dict(name='captcha', type='text')))
            
            captcha = captcha.getvalue()

        form = Form(action=self.urls['join'], submit='Join', post_html=captcha)
        form.add_element('textfield', 'Name')
        form.add_password_confirmation()
        form.add_element('hidden', 'key', value=str(key))
        return form

    def join(self, errors=None, wrap=False):
        """join div or page if wrap"""
        form = self.join_form()
        retval = markup.div(form(errors))
        if wrap:
            pagetitle = title = 'join'
            if self.site:
                pagetitle = '%s - %s' % (title, self.site)
            if self.captcha:
                errors = errors or {}
                captcha_err = errors.get('CAPTCHA', '')
                if captcha_err:
                    captcha_err = markup.p(markup.em(captcha_err),
                                           **{'class': 'error'})
            retval = markup.wrap(markup.h1(title.title()) + captcha_err + retval,
                                 pagetitle=pagetitle)
        return retval

    def join_post(self):
        """handle a join POST request"""
        form = self.join_form()
        errors = form.validate(self.request.POST)

        # validate captcha
        if CAPTCHA:
            key = self.request.POST.get('key')
            try:
                key = int(key)
            except ValueError:
                key = None
            if not key:
                errors['CAPTCHA'] = 'Please type the funky looking word'
            word = self.keys.pop(key, None)
            if not word:
                errors['CAPTCHA'] = 'Please type the funky looking word'
            if word != self.request.POST.get('captcha','').lower():
                errors['CAPTCHA'] = 'Sorry, you typed the wrong word'
        
        name = self.request.POST.get('Name', '')
        if not name:
            if not errors.has_key('Name'):
                errors['Name'] = []
            errors['Name'].append('Please enter a user name')
        if name in self.passwords():
            if not errors.has_key('Name'):
                errors['Name'] = []
            errors['Name'].append('The name %s is already taken' % name)

        if not errors: # create a new user
            self.newuser(name,
                         self.hash(name, self.request.POST['Password']))
            self.request.environ['REMOTE_USER'] = name # login the new user
            self.request.environ['paste.auth_tkt.set_user'](name)
        
        return errors

class BitsyAuth(object):
    """outer middleware for auth;  does the cookie handling and wrapping"""
    
    def __init__(self, app, global_conf, passwords, newuser, site='', secret='secret'):
        self.app = app
        self.path = '/logout'
        self.cookie = '__ac'
        auth = BitsyAuthInnerWare(app, passwords=passwords, newuser=newuser, site=site)

        # paste.auth.cookie
        #        self.cookie_handler = cookie.middleware(auth, cookie_name=self.cookie, timeout=90) # minutes

        # paste.auth.auth_tkt
        self.cookie_handler = auth_tkt.make_auth_tkt_middleware(auth, global_conf, secret, cookie_name=self.cookie, logout_path='/logout')

    def __call__(self, environ, start_response):
        if environ['PATH_INFO'] == '/logout':
            pass        
        return self.cookie_handler(environ, start_response)

    def logout(self, environ):
        req = Request(environ)
        keys = [ 'REMOTE_USER' ]
        #        keys = [ 'REMOTE_USER', 'AUTH_TYPE', 'paste.auth.cookie', 'paste.cookies', 'HTTP_COOKIE' ]  # XXX zealous kill
        for key in keys:
            req.environ.pop(key, None)

        body = '<html><head><title>logout</title></head><body>logout</body></html>'
        res = Response(content_type='text/html', body=body)
        res.content_length = len(res.body)
        req.cookies.pop(self.cookie, None)
        res.delete_cookie(self.cookie)
        res.unset_cookie(self.cookie)
        return res(environ, start_response)
        



import os
import settings
import shutil
from bitsyblog import roles
from webob import exc

class BitsyUser(object):
    """interface class for a bitsyblog user"""
    settings = {} 
    def __init__(self, name, password=''):
        self.name = name
        self.password = password

    # user behaves like a dictionary of settings

    def __getitem__(self, key):
        return self.settings[key]

    def get(self, key, default=None):
        return self.settings.get(key, default)

class BitsyUsers(object):
    """abstract class for bitsyblog user management"""

    def __iter__(self): return self.users()

    def __contains__(self, user):
        return user in self.users()

    def __getitem__(self, user):
        """return a user"""
        if user not in self.users():
            raise KeyError
        user = BitsyUser(user, self.password(user))
        user.settings = self.settings(user.name)
        return user

    def passwords(self):
        """returns a dictionary of { user: password }"""
        passwords = {}
        for user in self.users():            
            passwords[user] = self.password(user)
        return passwords

    ### interface methods to be specified by the child class

    def new(self, name, password):
        """create a new user"""

    def users(self):
        """returns the ids of all users (generator)"""

    def password(self, user):
        """return the password for the user"""

    def settings(self, user):
        """get user settings"""

    def write_settings(self, user, **kw):
        """set attributes on a user"""


class FilespaceUsers(BitsyUsers):
    """users that live on the filesystem"""

    def __init__(self, directory):
        BitsyUsers.__init__(self)
        self.directory = directory # directory to store user information in

    def home(self, user, *path):
        return os.path.join(self.directory, user, *path)

    def pw_file(self, user):
        return self.home(user, '.password')

    def preferences_file(self, user):
        return self.home(user, 'preferences.txt')

    def css(self, user, default):
        css_dir = self.home(user, 'css')
        css_files = [ i for i in os.listdir(css_dir) if i.endswith('.css') ]
        if default:
            default = '%s.css' % default
            try:
                index = css_files.index(default)
                css_files.insert(0, css_files.pop(index))
            except ValueError:
                pass
        return [ dict(filename=i, name=i.rsplit('.css',1)[0],
                      css=file(os.path.join(css_dir, i)).read())
                 for i in css_files ]

    ### interfaces for BitsyUsers

    def new(self, name, password):
        """create a new user account"""

        if name in self.users():
            raise exc.HTTPForbidden("The name %s is already taken" % name).exception

        # characters forbidden in user name
        forbidden = ' |<>./?,'
        urls = [ 'join', 'login', 'logout', 'feed' ]
        if [ i for i in forbidden if i in name ]:
            raise exc.HTTPForbidden("The name '%s' contains forbidden characters [%s]" % (user, forbidden)).exception
        if name in urls:
            raise exc.HTTPForbidden("The name '%s' is already used for a url"  % user).exception

        home = self.home(name)
        os.mkdir(home)
        pw_file = file(self.pw_file(name), 'w')
        print >> pw_file, password

        # setup entries structure for blog
        entries = os.path.join(home, 'entries') 
        os.mkdir(entries)
        for setting in roles['author']:
            os.mkdir(os.path.join(entries, setting))

        # setup user CSS
        css_dir = os.path.join(home, 'css') 
        os.mkdir(css_dir)
        shutil.copyfile(os.path.join(self.directory, 'site.css'),
                        os.path.join(css_dir, 'default.css'))


    def users(self):
        ignores = set(['.svn'])
        for user in os.listdir(self.directory):
            # ensure integrity of user folder
            if user in ignores:
                continue
            if os.path.isdir(os.path.join(self.directory, user)):
                yield user

    def password(self, user):
        pw_file = '.password' # name of the password file on the filesystem
        password = self.home(user, pw_file)
        if os.path.exists(password):
            return file(password).read().strip()
        return  '' # unspecified password

    def settings(self, name):
        """returns a dictionary of user preferences from a file"""
        filename = self.home(name, 'preferences.txt')
        prefs = {}
        if os.path.exists(filename):
            prefs = file(filename).read().split('\n')
            prefs = [ i for i in prefs if i.strip() ]
            prefs = [ [ j.strip() for j in i.split(':', 1) ] 
                      for i in prefs if ':' in i]
            prefs = dict(prefs)

        # assemble friends from a list
        friends = prefs.get('Friends') # can see secret blog posts
        if friends:
            prefs['Friends'] = friends.split(', ')
        else:
            prefs['Friends'] = []

        # CSS files
        prefs['CSS'] = self.css(name, prefs.get('Stylesheet'))
        
        return prefs

    def write_settings(self, name, **kw):
        """write user settings to disk"""
        
        # generic stuff; could factor out
        
        newsettings = {}
        errors = {}

        for setting in settings.form:
            if kw.has_key(setting.name):
                try:
                    setting.set(kw[setting.name])
                    newsettings[setting.name] = setting.value
                except settings.InvalidSettingError, e:
                    errors[setting.name] = str(e)

        if errors:
            return errors

        # this makes the function depend on implemention
        # i don't like this
        new_css = newsettings.pop('CSS file')
        if new_css:
            filename = new_css['filename']
            css_file = file(self.home(name, 'css', filename), 'w')
            print >> css_file, new_css['css']
            newsettings['CSS'] = filename.rsplit('.css', 1)[0]
                
        prefs = self.settings(name)
        prefs.update(newsettings)

        # write the preferences to a file
        preferences = file(self.preferences_file(name), 'w')
        for key, value in prefs.items():
            print >> preferences, '%s: %s' % ( key, value )



from setuptools import setup, find_packages
import sys, os

version = '0.7.2'

try:
    description = file('README.txt').read()
except IOError:
    description = """
Meet bitsyblog.  Posting is done with a POST request, so while you can use
a web form to do this, its just as easy to use curl, urllib, or anything else 
to post.
"""

setup(name='bitsyblog',
      version=version,
      description="a tiny tiny blog",
      long_description=description,
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='blog',
      author='Jeff Hammel',
      author_email='jhammel@openplans.org',
      url='http://bitsyblog.biz',
      license='GPL',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
          'WebOb',
          'Paste',
          'PasteScript',
          'dateutil',
          'markup',
          'skimpygimpy',
          'lxml'
      ],
      dependency_links=[ 'https://svn.openplans.org/svn/standalone/markup#egg=markup',
                         'http://svn.pythonpaste.org/Paste/trunk#egg=Paste',
                         'http://downloads.sourceforge.net/skimpygimpy/skimpyGimpy_1_3.zip#egg=skimpygimpy'
                         ],
      entry_points="""
      # -*- Entry points: -*-
      [paste.app_factory]
      main = bitsyblog.factory:factory
      """,
      )



"""blog interfaces to data for bitsy"""

import os
import utils

from cStringIO import StringIO
from glob import glob

class BlogEntry(object):
    """interface class for a blog entry"""
    def __init__(self, date, body, privacy):
        self.date = date
        self.body = body
        self.privacy = privacy

    def snippet(self, characters=80):
        if characters:
            if len(self.body) > characters:
                text = ' '.join(self.body[:characters].split()[:-1])
                if text:
                    return '%s ...' % text
            else:
                return self.body
        return ''
                

class Blog(object):
    """abstract class for a users' blog"""

    def __call__(self, user, permissions=('public',), number=None):
        return self.blog(user, permissions, number=number)

    def blog(self, user, permissions, number=None):
        """
        return the user's blog sorted in reverse date order
        if number is None, the entire blog is returned
        """

    def entry(self, user, datestamp, permissions):
        """
        return a single blog entry with the given datestamp:
        'YYYYMMDDHHMMSS'
        """

    def entries(self, user, permissions, year=None, month=None, day=None):
        """return entries by date"""

    def post(self, user, date, text, privacy):
        """post a new blog entry"""
        

class FileBlog(Blog):
    """a blog that lives on the filesystem"""

    def __init__(self, directory):
        self.directory = directory

    def location(self, user, permission, *path):
        """returns which directory files are in based on permission"""
        return os.path.join(self.directory, user, 'entries', permission, *path)

    def body(self, user, datestamp, permission):
        return file(self.location(user, permission, datestamp)).read()

    def get_entry(self, user, datestamp, permission):
        return BlogEntry(utils.date(datestamp), 
                         self.body(user, datestamp, permission),
                         permission)

    ### interfaces from Blog

    def blog(self, user, permissions, number=None):        
        entries = []
        for permission in permissions:
            entries.extend([ (entry, permission) 
                             for entry in os.listdir(self.location(user, permission)) ])
        entries.sort(key=lambda x: x[0], reverse=True)
        
        if number is not None:
            entries = entries[:number]

        return [ self.get_entry(user, x[0], x[1]) for x in entries ]

    def entry(self, user, datestamp, permissions):
        for permission in permissions:
            filename = self.location(user, permission, datestamp)
            if os.path.exists(filename):
                return self.get_entry(user, datestamp, permission)

    def entries(self, user, permissions, year=None, month=None, day=None):

        # build a file glob expression
        dateargs = [ year, month, day ]
        glob_expr = StringIO()
        for index in range(len(dateargs)):
            value = dateargs[index]
            if value is None:
                break
            length = len(utils.timeformat[index])
            print >> glob_expr, '0*d' % (length, int(value))
        
        while index < len(utils.timeformat):
            print >> glob_expr, '[0-9]' * len(utils.timeformat[index])
            index += 1
        glob_expr = glob_expr.getvalue()

        # get the blog entries
        entries = []
        for permission in permissions:
            entries.extend([ (os.path.split(entry)[-1], permission) 
                             for entry in glob(os.path.join(self.location(user, permission), glob_expr)) ])
        entries.sort(key=lambda x: x[0], reverse=True)
        return [ self.get_entry(user, x[0], x[1]) for x in entries ]

    def post(self, user, datestamp, body, privacy):
        blog = file(self.location(user, privacy, datestamp), 'w')
        print >> blog, body



"""user settings / preferences"""

import parser # dateutils.parser with mods to retain strftime format
import urllib2
import utils

class InvalidSettingError(Exception):
    """error when trying to validate a setting"""

class Setting(object):
    def __init__(self, name, value=None):
        self.name = name
        self.value = value
        self.error_message = 'Invalid value for %s' % name

    def validator(self, value):
        return True

    def set(self, value):
        if not self.validator(value):
            raise InvalidSettingError(self.error_message)
        self.value = value

class DateFormat(Setting):
    def __init__(self):
        Setting.__init__(self, 'Date format')

    def set(self, format):
        value = parser.parser()._parse(format)
        if value:
            self.value = value.format
        else:
            raise InvalidSettingError('unrecognized date format: %s' % format)

    
class CSSFile(Setting):
    def __init__(self):
        Setting.__init__(self, 'CSS file')

    def set(self, value):
        if not hasattr(value, 'file'):
            return True # blank set: don't do anything
        css = value.file.read()
        try:
            validcss = utils.validate_css(css)
        except urllib2.URLError:
            raise InvalidSettingError('Could not validate CSS (sorry!)')
        if not validcss:
            raise InvalidSettingError('%s is not valid css' % filename)
        filename = value.filename
        if not filename.endswith('.css'):
            filename = '%s.css' % filename
        self.value = dict(filename=filename, css=css)

user = [ DateFormat(),
         Setting('Subject'),
         Setting('Friends'),
         Setting('Stylesheet'),
         ]

form = user[:]
form += [ CSSFile() ]



"""utlity functions for bitsyblog"""

import datetime
import os
import urllib
import urllib2

timeformat = ( 'YYYY', 'MM', 'DD', 'HH', 'MM', 'SS' )

def validate_css(css):
    """use a webservice to determine if the argument is valid css"""    
    url = 'http://jigsaw.w3.org/css-validator/validator?text=%s'
    url = url % urllib.quote_plus(css)
    foo = urllib2.urlopen(url)
    text = foo.read()
    return not 'We found the following errors' in text

def date(datestamp):
    datestamp = os.path.split(datestamp)[-1]
    retval = []
    for i in timeformat:
        retval.append(int(datestamp[:len(i)]))
        datestamp = datestamp[len(i):]
    return datetime.datetime(*retval)

